#!/usr/bin/env python3

_license = """
big
Copyright 2022 Larry Hastings
All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import io
import unittest

stats = {
    "failures": 0,
    "errors": 0,
    "expected failures": 0,
    "unexpected successes": 0,
    "skipped": 0,
    }

attribute_names = {
    "failures": "failures",
    "errors": "errors",
    "expected failures": "expectedFailures",
    "unexpected successes": "unexpectedSuccesses",
    "skipped": "skipped",
    }

def run(name, module, permutations=None):
    if name:
        print(f"Testing {name}...")

    # this is a lot of work to suppress the "\nOK"!
    sio = io.StringIO()
    runner = unittest.TextTestRunner(stream=sio)
    t = unittest.main(module=module, exit=False, testRunner=runner)
    result = t.result
    for name in stats:
        value = getattr(result, attribute_names[name], ())
        stats[name] += len(value)

    for line in sio.getvalue().split("\n"):
        if not line.startswith("Ran"):
            print(line)
            continue

        if permutations:
            fields = line.split()
            assert fields[2] == "tests"
            fields[2] = f"tests, with {permutations()} total permutations,"
            line = " ".join(fields)
        print(line)
        print()
        # prevent printing the  \n and OK
        break

def finish():
    if not (stats['failures'] or stats['errors']):
        result = "OK"
    else: # pragma: no cover
        result = "FAILED"

    fields = [f"{name}={value}" for name, value in stats.items() if value]
    if fields: # pragma: no cover
        addendum = ", ".join(fields)
        result = f"{result} ({addendum})"
    print(result)



#!/usr/bin/env python3

_license = """
big
Copyright 2022 Larry Hastings
All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import enum
import functools
import itertools
from itertools import zip_longest
from .itertools import PushbackIterator
import operator
import re
import struct
import sys


__all__ = []

def _export_name(s):
    __all__.append(s)

def _export(o):
    _export_name(o.__name__)
    return o


@_export
def re_partition(s, pattern, count=1, *, flags=0, reverse=False):
    """
    Like str.partition, but pattern is matched as a regular expression.

    s can be a string or a bytes object.

    pattern can be a string, bytes, or an re.Pattern object.

    s and pattern (or pattern.pattern) must be the same type.

    If pattern is found in s, returns a tuple
        (before, match, after)
    where before is the text before the match,
    match is the re.Match object resulting from the match, and
    after is the text after the match.

    If pattern appears in s multiple times,
    re_partition will match against the first (leftmost)
    appearance.

    If pattern is not found in s, returns a tuple
        (s, None, '')
    where the empty string is str or bytes as appropriate.

    Passing in an explicit "count" lets you control how many times
    re_partition partitions the string.  re_partition will always
    return a tuple containing (2*count)+1 elements, and
    odd-numbered elements will be either re.Match objects or None.
    Passing in a count of 0 will always return a tuple containing s.

    If pattern is a string or bytes, flags is passed in
    as the flags argument to re.compile.

    If reverse is true, partitions starting at the right,
    like re_rpartition.
    """
    if reverse:
        return re_rpartition(s, pattern, count, flags=flags)
    if not isinstance(pattern, re.Pattern):
        pattern = re.compile(pattern, flags=flags)
    if not (isinstance(s, (str, bytes)) and
        (type(s) == type(pattern.pattern))):
        raise ValueError("s must be str or int, and string used in pattern must be the same type")
    as_bytes = isinstance(s, bytes)
    empty = b'' if as_bytes else ''

    if count < 0:
        raise ValueError("count must be >= 0")

    result = [s]
    if count == 0:
        return result

    if count == 1:
        # much cheaper for the general case
        match = pattern.search(s)
        if match:
            matches = [match]
        else:
            matches = ()
    else:
        matches = list(pattern.finditer(s))
        matches.reverse()

    for _ in range(count):
        if not matches:
            result.extend((None, empty))
            continue

        match = matches.pop()
        s = result.pop()

        before, separator, after = s.partition(match.group(0))
        result.extend((before, match, after))

    return tuple(result)


@_export
def re_rpartition(s, pattern, count=1, *, flags=0):
    """
    Like str.rpartition, but pattern is matched as a regular expression.

    s can be a string or a bytes object.

    pattern can be a string, bytes, or an re.Pattern object.

    s and pattern (or pattern.pattern) must be the same type.

    If pattern is found in s, returns a tuple
        (before, match, after)
    where before is the text before the match,
    match is the re.Match object resulting from the match, and
    after is the text after the match.

    If pattern appears in s multiple times,
    re_partition will match against the last (rightmost)
    appearance.

    If pattern is not found in s, returns a tuple
        ('', None, s)
    where the empty string is str or bytes as appropriate.

    Passing in an explicit "count" lets you control how many times
    re_rpartition partitions the string.  re_rpartition will always
    return a tuple containing (2*count)+1 elements, and
    odd-numbered elements will be either re.Match objects or None.
    Passing in a count of 0 will always return a tuple containing s.

    If pattern is a string, flags is passed in
    as the flags argument to re.compile.
    """
    if not isinstance(pattern, re.Pattern):
        pattern = re.compile(pattern, flags=flags)
    if not (isinstance(s, (str, bytes)) and
        (type(s) == type(pattern.pattern))):
        raise ValueError("s must be str or int, and string used in pattern must be the same type")

    if count < 0:
        raise ValueError("count must be >= 0")

    result = [s]
    if count == 0:
        return result

    as_bytes = not isinstance(s, str)
    empty = b'' if as_bytes else ''
    matches = list(pattern.finditer(s))

    for i in range(count):
        if not matches:
            r = [empty, None]
        else:
            match = matches.pop()
            s = result.pop(0)
            before, separator, after = list(s.rpartition(match.group(0)))
            r = [before, match, after]
        r.extend(result)
        result = r
    return tuple(result)


# a list of all unicode whitespace characters known to Python
_export_name('whitespace')
whitespace = (
    '\t'    , #     9 0x0009 - tab
    '\n'    , #    10 0x000a - newline
    '\x0b'  , #    11 0x000b - vertical tab
    '\x0c'  , #    12 0x000c - form feed
    '\r'    , #    13 0x000d - carriage return
    '\x1c'  , #    28 0x001c - file separator
    '\x1d'  , #    29 0x001d - group separator
    '\x1e'  , #    30 0x001e - record separator
    '\x1f'  , #    31 0x001f - unit separator
    ' '     , #    32 0x0020 - space
    '\x85'  , #   133 0x0085 - next line
    '\xa0'  , #   160 0x00a0 - non-breaking space
    '\u1680', #  5760 0x1680 - ogham space mark
    '\u2000', #  8192 0x2000 - en quad
    '\u2001', #  8193 0x2001 - em quad
    '\u2002', #  8194 0x2002 - en space
    '\u2003', #  8195 0x2003 - em space
    '\u2004', #  8196 0x2004 - three-per-em space
    '\u2005', #  8197 0x2005 - four-per-em space
    '\u2006', #  8198 0x2006 - six-per-em space
    '\u2007', #  8199 0x2007 - figure space
    '\u2008', #  8200 0x2008 - punctuation space
    '\u2009', #  8201 0x2009 - thin space
    '\u200a', #  8202 0x200a - hair space
    '\u2028', #  8232 0x2028 - line separator
    '\u2029', #  8233 0x2029 - paragraph separator
    '\u202f', #  8239 0x202f - narrow no-break space
    '\u205f', #  8287 0x205f - medium mathematical space
    '\u3000', # 12288 0x3000 - ideographic space

    '\r\n'  , # the classic DOS newline sequence
    )

# this omits the DOS convention '\r\n'
_export_name('whitespace_without_dos')
whitespace_without_dos = tuple(s for s in whitespace if s != '\r\n')

_export_name('newlines')
newlines = (
    '\n'    , #   10 0x000a - newline
    '\x0b'  , #   11 0x000b - vertical tab
    '\x0c'  , #   12 0x000c - form feed
    '\r'    , #   13 0x000d - carriage return
    '\x1c'  , #   28 0x001c - file separator
    '\x1d'  , #   29 0x001d - group separator
    '\x1e'  , #   30 0x001e - record separator
    '\x85'  , #  133 0x0085 - next line
    '\u2028', # 8232 0x2028 - line separator
    '\u2029', # 8233 0x2029 - paragraph separator

    '\r\n'  , # the classic DOS newline sequence

    # what about '\n\r'?
    # sorry, Acorn and RISC OS users, you'll have to add this yourselves.
    # I'm worried it would cause bugs with a malformed DOS newline.
    )

# this omits the DOS convention '\r\n'
_export_name('newlines_without_dos')
newlines_without_dos = tuple(s for s in newlines if s != '\r\n')

def _cheap_encode_iterable_of_strings(iterable, encoding='ascii'):
    a = []
    # all I'm really doing with the try/accept
    # block is stopping the loop for the ascii
    # encoding once we hit characters > 127.
    # since we're sorted, by the time we hit it,
    # we'll have already handled all the ascii
    # characters (if we're encoding for ascii).
    try:
        for c in iterable:
            a.append(c.encode(encoding))
    except UnicodeEncodeError:
        pass
    return type(iterable)(a)

# use these with multisplit if your bytes strings are encoded as ascii.
_export_name('ascii_whitespace')
ascii_whitespace = _cheap_encode_iterable_of_strings(whitespace, "ascii")
_export_name('ascii_whitespace_without_dos')
ascii_whitespace_without_dos = tuple(s for s in ascii_whitespace if s != b'\r\n')
_export_name('ascii_newlines')
ascii_newlines   = _cheap_encode_iterable_of_strings(newlines,   "ascii")
_export_name('ascii_newlines_without_dos')
ascii_newlines_without_dos = tuple(s for s in ascii_newlines if s != b'\r\n')

# use these with multisplit if your bytes strings are encoded as utf-8.
_export_name('utf8_whitespace')
utf8_whitespace = _cheap_encode_iterable_of_strings(whitespace, "utf-8")
_export_name('utf8_whitespace_without_dos')
utf8_whitespace_without_dos = tuple(s for s in utf8_whitespace if s != b'\r\n')
_export_name('utf8_newlines')
utf8_newlines   = _cheap_encode_iterable_of_strings(newlines,   "utf-8")
_export_name('utf8_newlines_without_dos')
utf8_newlines_without_dos =  tuple(s for s in utf8_newlines if s != b'\r\n')

def _multisplit_reversed(o):
    if isinstance(o, str):
        if len(o) <= 1:
            return o
        return "".join(reversed(o))
    if isinstance(o, bytes):
        if len(o) <= 1:
            return o
        return b"".join(o[i:i+1] for i in range(len(o)-1, -1, -1))
    t = type(o)
    assert t in (list, tuple, set), f"{o=} {t=}"
    return t(_multisplit_reversed(p) for p in o)


# _reversed_builtin_separators precalculates the reversed versions
# of the builtin separators.  we use the reversed versions when
# reverse=True.  this is a minor speed optimization, but also it
# helps a lot with the lrucache for _separators_to_re.
#
# we test that these cached versions are correct in tests/test_text.py.
#
_reversed_utf8_whitespace_without_dos = _multisplit_reversed(utf8_whitespace_without_dos)
_reversed_utf8_newlines_without_dos =   _multisplit_reversed(utf8_newlines_without_dos)

_reversed_builtin_separators = {
    whitespace: whitespace_without_dos + ('\n\r',),
    whitespace_without_dos: whitespace_without_dos,

    newlines: newlines_without_dos + ("\n\r",),
    newlines_without_dos: newlines_without_dos,

    ascii_whitespace: ascii_whitespace_without_dos + (b"\n\r",),
    ascii_whitespace_without_dos: ascii_whitespace_without_dos,

    ascii_newlines: ascii_newlines_without_dos + (b"\n\r",),
    ascii_newlines_without_dos: ascii_newlines_without_dos,

    utf8_whitespace: _reversed_utf8_whitespace_without_dos + (b"\n\r",),
    utf8_whitespace_without_dos: _reversed_utf8_whitespace_without_dos,

    utf8_newlines: _reversed_utf8_newlines_without_dos + (b"\n\r",),
    utf8_newlines_without_dos: _reversed_utf8_newlines_without_dos,
    }


def _re_quote(s):
    # don't bother escaping whitespace.
    # re.escape escapes whitespace because of VERBOSE mode,
    # which we're not using.  (escaping the whitespace doesn't
    # hurt anything really, but it makes the patterns harder
    # to read for us humans.)
    if not s.isspace():
        s = re.escape(s)
    if len(s) > 1:
        if isinstance(s, bytes):
            s = b"(?:" + s + b")"
        else:
            s = f"(?:{s})"
    return s


@functools.lru_cache(re._MAXCACHE)
def __separators_to_re(separators, as_bytes, separate=False, keep=False):
    if as_bytes:
        pipe = b'|'
        separate_start = b'(?:'
        separate_end = b')+'
        keep_start = b'('
        keep_end = b')'
    else:
        pipe = '|'
        separate_start = '(?:'
        separate_end = ')+'
        keep_start = '('
        keep_end = ')'

    # sort longer separator strings earlier.
    # re processes | operator from left-to-right,
    # so you want to match against longer strings first.
    separators = list(separators)
    separators.sort(key=lambda o: -len(o))
    pattern = pipe.join(_re_quote(o) for o in separators)
    # print(f"  P1 {pattern!r}")
    if not separate:
        pattern = separate_start + pattern + separate_end
    # print(f"  P2 {pattern!r}")
    if keep:
        pattern = keep_start + pattern + keep_end
    # print(f"  P3 {pattern!r}")
    return pattern

def _separators_to_re(separators, as_bytes, separate=False, keep=False):
    try:
        hash(separators)
    except TypeError:
        separators = tuple(separators)
    return __separators_to_re(separators, bool(as_bytes), separate=bool(separate), keep=bool(keep))



@_export
def multistrip(s, separators, left=True, right=True):
    """
    Like str.strip, but supports stripping multiple strings.

    Strips from the string "s" all leading and trailing
    instances of strings found in "separators".

    s should be str or bytes.
    separators should be an iterable of either str or bytes
    objects matching the type of s.

    If left is a true value, strips all leading separators
    from s.

    If right is a true value, strips all trailing separators
    from s.

    Processing always stops at the first character that
    doesn't match one of the separators.

    Returns a copy of s with the leading and/or trailing
    separators stripped.  (If left and right are both
    false, returns s unchanged.)
    """
    if not (left or right):
        return s
    as_bytes = isinstance(s, bytes)
    if as_bytes:
        head = b'^'
        tail = b'$'
        if isinstance(separators, bytes):
            # not iterable of bytes, literally a bytes string.
            # split it ourselves.  otherwise, _separators_to_re will
            # iterate over it, and... oops! it gets bytes!
            separators = tuple(separators[i:i+1] for i in range(len(separators)))
    else:
        head = '^'
        tail = '$'
    pattern = _separators_to_re(separators, as_bytes, separate=False, keep=False)

    start = 0
    end = len(s)
    if left:
        left_match = re.match(head + pattern, s)
        if left_match:
            start = left_match.end(0)
            s = s[start:]
    if right:
        right_match = re.search(pattern + tail, s)
        if right_match:
            end = right_match.start(0)
            s = s[:end]
    return s


# for keep
AS_PAIRS="AS_PAIRS"
_export_name(AS_PAIRS)
ALTERNATING="ALTERNATING"
_export_name(ALTERNATING)

# for strip
LEFT = "LEFT"
_export_name(LEFT)
RIGHT = "RIGHT"
_export_name(RIGHT)
PROGRESSIVE = "PROGRESSIVE"
_export_name(PROGRESSIVE)

@_export
def multisplit(s, separators=None, *,
    keep=False,
    maxsplit=-1,
    reverse=False,
    separate=False,
    strip=False,
    ):
    """
    Splits strings like str.split, but with multiple separators and options.

    s can be str or bytes.

    separators should be an iterable of str or bytes, matching s.

    Returns an iterator yielding the strings split from s.  If keep
    is true (or ALTERNATING), and strip is false, joining these strings
    together will recreate s.

    multisplit is greedy; if two (or more) separator strings match at a
    particular index in "s", it splits using the largest one.

    "keep" indicates whether or not multisplit should keep the separator
    strings.  It supports four values:
        false (the default)
            Discard the separators.
        true (apart from ALTERNATING and AS_PAIRS)
            Append the separators to the end of the split strings.
            You can recreate the original string by passing the
            list returned in to ''.join.
        ALTERNATING
            Yield alternating strings in the output: strings consisting
            of separators, alternating with strings consisting of
            non-separators.  If "separate" is true, separator strings
            will contain exactly one separator, and non-separator strings
            may be empty; if "separate" is false, separator strings will
            contain one or more separators, and non-separator strings
            will never be empty.
            You can recreate the original string by passing the
            list returned in to ''.join.
        AS_PAIRS
            Yield 2-tuples containing a non-separator string and its
            subsequent separator string.  Either string may be empty;
            the separator string in the last 2-tuple will always be
            empty, and if `s` ends with a separator string, *both*
            strings in the final 2-tuple will be empty.

    "separate" indicates whether multisplit should consider adjacent
    separator strings in s as one separator or as multiple separators
    each separated by a zero-length string.  It supports two values:
        false (the default)
            Multiple adjacent separators should be considered one
            separator.
        true
            Don't group separators together.  Each separator should
            split the string individually, even if there are no
            characters between two separators.

    "strip" indicates whether multisplit should strip separators from
    the beginning and/or end of s, a la multistrip.  It supports
    six values:
        false (the default)
            Don't strip separators from the beginning or end of s.
        true (apart from LEFT, RIGHT, and PROGRESSIVE)
            Strip separators from the beginning and end of s
            (a la str.strip).
        LEFT
            Strip separators only from the beginning of s
            (a la str.lstrip).
        RIGHT
            Strip separators only from the end of s
            (a la str.rstrip).
        PROGRESSIVE
            Strip from the beginning and end of s, unless maxsplit
            is nonzero and the entire string is not split.  If
            splitting stops due to maxsplit before the entire string
            is split, and reverse is false, don't strip the end of
            the string. If splitting stops due to maxsplit before
            the entire string is split, and reverse is true, don't
            strip the beginning of the string.  (This is how str.strip
            and str.rstrip behave when sep=None.)

    "maxsplit" should be either an integer or None.  If maxsplit is an
    integer greater than -1, multisplit will split s no more than
    maxsplit times.

    "reverse" controls whether multisplit splits starting from the
    beginning or from the end of the string.  It supports two values:
        false (the default)
            Start splitting from the beginning of the string
            and scanning right.
        true
            Start splitting from the end of the string and
            scanning left.
    Splitting from the end of the string and scanning left has two
    effects.  First, if maxsplit is a number greater than 0,
    the splits will start at the end of the string rather than
    the beginning.  Second, if there are overlapping instances of
    separators in the string, multisplit will prefer the rightmost
    separator rather than the left.  For example:
        multisplit(" x x ", (" x ",), keep=big.ALTERNATING)
    will split on the leftmost instance of " x ", yielding
        "", " x ", "x "
    whereas
        multisplit(" x x ", (" x ",), keep=big.ALTERNATING, reverse=True)
    will split on the rightmost instance of " x ", yielding
        " x", " x ", ""
    """
    as_bytes = isinstance(s, bytes)
    if separators is None:
        separators = ascii_whitespace if as_bytes else whitespace
    elif not separators:
        raise ValueError("illegal separators")

    if as_bytes:
        if isinstance(separators, bytes):
            # not iterable of bytes, literally a bytes string.
            # split it ourselves.
            separators = tuple(separators[i:i+1] for i in range(len(separators)))
        separators_type = bytes
        empty = b''
    else:
        separators_type = str
        empty = ''
    for o in separators:
        if not isinstance(o, separators_type):
            raise ValueError("separators must be an iterable of objects the same type as s")

    separators_to_re_keep = keep

    if strip:
        if strip == PROGRESSIVE:
            if maxsplit == -1:
                strip = left = right = True
            else:
                left = not reverse
                right = reverse
                separators_to_re_keep = True
        else:
            left = strip != RIGHT
            right = strip != LEFT
        s = multistrip(s, separators, left=left, right=right)
        if not s:
            # oops! all separators!
            # this will make us exit early, just a few lines down from here.
            maxsplit = 0

    if maxsplit == None:
        maxsplit = -1
    if maxsplit == 0:
        if keep == ALTERNATING:
            yield s
        elif keep == AS_PAIRS:
            yield (s, empty)
        else:
            yield s
        return

    # convert maxsplit for use with re.split.
    #
    # re.split interprets maxsplit slightly differently:
    #   its maxsplit==0 means "allow all splits".
    #   its maxsplit==1 means "only allow one split".
    #
    # (re.split doesn't have a way to express
    #  "don't split" with its maxsplit parameter,
    #  which is why we handled it already.)
    re_split_maxsplit = maxsplit if maxsplit != -1 else 0

    if reverse:
        # if reverse is true, when separators overlap,
        # we need to prefer the rightmost one rather than
        # the leftmost one.  how do we do *that*?
        # Eric Smith had the brainstorm: reverse the string
        # and the separators, split, and reverse the output
        # and the strings in the output.
        s = _multisplit_reversed(s)
        s2 = _reversed_builtin_separators.get(separators, None)
        if s2 != None:
            separators = s2
        else:
            separators = _multisplit_reversed(separators)

    pattern = _separators_to_re(separators, as_bytes, keep=separators_to_re_keep, separate=separate)
    # print("PATTERN", pattern, f"{separators_to_re_keep=} {separate=}")

    l = re.split(pattern, s, re_split_maxsplit)
    assert l
    # print("S", repr(s), "L", l, f"{re_split_maxsplit=}")

    if strip == PROGRESSIVE:
        # l alternates nonsep and sep strings.
        # it's always an odd length, starting and ending with nonsep.
        length = len(l)
        assert length & 1

        desired_length = 1 + (2*maxsplit)

        # dang! this is complicated!
        # maxsplit has to extend *past* the last nonsep
        # for us to strip on the far side.
        #  ' a b c   '.split(None, maxsplit=2) => ['a', 'b', 'c   ']
        #  ' a b c   '.split(None, maxsplit=3) => ['a', 'b', 'c']
        for i in range(length - 1, 0, -2):
            nonsep = l[i]
            if nonsep:
                last_non_empty_nonsep = i
                break
        else:
            last_non_empty_nonsep = 0

        if desired_length > (last_non_empty_nonsep + 2):
            # strip!
            l = l[:last_non_empty_nonsep + 1]
            desired_length = length = last_non_empty_nonsep

        if not keep:
            for i in range(len(l) - 2, 0, -2):
                del l[i]

    if reverse:
        l = _multisplit_reversed(l)
        l.reverse()

    if not keep:
        for o in l:
            yield o
        return

    # from here on out, we're 'keep'-ing the separator strings.
    # (we're returning the separator strings in one form or another.)

    if keep == ALTERNATING:
        for o in l:
            yield o
        return

    if (len(l) % 2) == 1:
        l.append(empty)

    previous = None
    for o in l:
        if previous is None:
            previous = o
            continue
        if keep == AS_PAIRS:
            yield (previous, o)
        else:
            yield previous + o
        previous = None


@_export
def multipartition(s, separators, count=1, *, reverse=False, separate=True):
    """
    Like str.partition, but supports partitioning based on multiple separator
    strings, and can partition more than once.

    "s" can be str or bytes.

    "separators" should be an iterable of str or bytes, matching "s".

    By default, if any of the strings in "separators" are found in "s",
    returns a tuple of three strings: the portion of "s" leading up to
    the earliest separator, the separator, and the portion of "s" after
    that separator.  Example:

    multipartition('aXbYz', ('X', 'Y')) => ('a', 'X', 'bYz')

    If none of the separators are found in the string, returns
    a tuple containing `s` unchanged followed by two empty strings.

    Passing in an explicit "count" lets you control how many times
    multipartition partitions the string.  multipartition will always
    return a tuple containing (2*count)+1 elements.  Passing in a
    count of 0 will always return a tuple containing s.

    If separate is true, multiple adjacent separator strings behave
    like one separator.  Example:

    multipartition('aXYbXYc', ('X', 'Y',), separate=True) => ('a', 'XY', 'bXYc')

    If reverse is true, multipartition behaves like str.rpartition.
    It partitions starting on the right, scanning backwards through
    s looking for separators.
    """
    if count < 0:
        raise ValueError("count must be positive")
    result = list(multisplit(s, separators,
        keep=ALTERNATING,
        reverse=reverse,
        separate=separate,
        strip=False,
        maxsplit=count))
    desired_length = (2 * count) + 1
    result_length = len(result)
    if result_length < desired_length:
        as_bytes = isinstance(s, bytes)
        if as_bytes:
            empty = b''
        else:
            empty = ''
        extension = [empty] * (desired_length - result_length)
        if reverse:
            result = extension + result
        else:
            result.extend(extension)
    return tuple(result)


_invalid_state = "_invalid_state"
_in_word = "_in_word"
_after_whitespace = "_after_whitespace"
_after_whitespace_then_apostrophe_or_double_quote = "_after_whitespace_then_apostrophe_or_double_quote"
_after_whitespace_then_D_or_O = "_after_whitespace_then_D_or_O"
_after_whitespace_then_D_or_O_then_apostrophe = "_after_whitespace_then_D_or_O_then_apostrophe"

@_export
def gently_title(s, *, apostrophes=None, double_quotes=None):
    """
    Uppercase the first character of every word in s.
    Leave the other letters alone.  s should be str or bytes.

    (For the purposes of this algorithm, words are
    any blob of non-whitespace characters.)

    Capitalize the letter after an apostrophe if
        a) the apostrophe is after whitespace
           (or is the first letter of the string), or
        b) if the apostrophe is after a letter O or D,
           and that O or D is after whitespace (or is
           the first letter of the string).  The O or D
           here will also be capitalized.
    Rule a) handles internally quoted strings:
            He Said 'No I Did Not'
        and contractions that start with an apostrophe
            'Twas The Night Before Christmas
    Rule b) handles certain Irish, French, and Italian
        names.
            Peter O'Toole
            Lord D'Arcy

    Capitalize the letter after a quote mark if
    the quote mark is after whitespace (or is the
    first letter of a string).

    A run of consecutive apostrophes and/or
    quote marks is considered one quote mark for
    the purposes of capitalization.

    If specified, apostrophes should be a str
    or bytes object containing characters that
    should be considered apostrophes.  If apostrophes
    is false, and s is bytes, apostrophes is set to "'".
    If apostrophes is false and s is str, apostrophes
    is set to a string containing these Unicode
    apostrophe code points:
        '‘’‚‛

    If specified, double_quotes should be a str
    or bytes object containing characters that
    should be considered double-quote characters.
    If double_quotes is false, and s is bytes,
    double_quotes is set to "'".
    If double_quotes is false, and s is str, double_quotes
    is set to a string containing these Unicode double quote
    code points:
        "“”„‟«»‹›
    """
    as_bytes = isinstance(s, bytes)
    if as_bytes:
        empty = b""
        if not apostrophes:
            apostrophes = b"'"
        if not double_quotes:
            double_quotes = b'"'
        d_and_o = b'DO'
        characters_in_s = (s[i:i+1] for i in range(len(s)))
    else:
        empty = ""
        if not apostrophes:
            apostrophes = "'‘’‚‛"
        if not double_quotes:
            double_quotes = '"“”„‟«»‹›'
        d_and_o = 'DO'
        characters_in_s = iter(s)

    # ooh, fancy!
    _is_apostrophe = apostrophes.__contains__
    _is_double_quote = double_quotes.__contains__
    _is_d_or_o = d_and_o.__contains__

    result = []
    state = _after_whitespace
    for c in characters_in_s:
        is_space = c.isspace()
        is_apostrophe = _is_apostrophe(c)
        is_double_quote = _is_double_quote(c)
        if state == _in_word:
            if is_space:
                state = _after_whitespace
        elif state == _after_whitespace:
            if not is_space:
                c = c.upper()
                if (is_apostrophe or is_double_quote):
                    state = _after_whitespace_then_apostrophe_or_double_quote
                elif _is_d_or_o(c):
                    state = _after_whitespace_then_D_or_O
                else:
                    state = _in_word
        elif state == _after_whitespace_then_apostrophe_or_double_quote:
            if not (is_apostrophe or is_double_quote):
                c = c.upper()
                state = _in_word
        elif state == _after_whitespace_then_D_or_O:
            if is_apostrophe:
                state = _after_whitespace_then_D_or_O_then_apostrophe
            else:
                state = _in_word
        elif state == _after_whitespace_then_D_or_O_then_apostrophe:
            c = c.upper()
            state = _in_word
        result.append(c)
    return empty.join(result)


@_export
def normalize_whitespace(s, separators=None, replacement=None):
    """
    Returns s, but with every run of consecutive
    separator characters turned into a replacement string.
    By default turns all runs of consecutive whitespace
    characters into a single space character.

    s may be str or bytes.
    separators should be an iterable of either str or bytes objects,
    matching s.
    replacement should be either a str or bytes object,
    also matching s, or None (the default).
    If replacement is None, normalize_whitespace will use
    a replacement string consisting of a single space character,
    either str or bytes as appropriate.

    Leading or trailing runs of separator characters will
    be replaced with the replacement string, e.g.:

       normalize_whitespace("   a    b   c") == " a b c".
    """
    as_bytes = isinstance(s, bytes)
    if as_bytes:
        empty = b''
        default_replacement = b' '
        default_separators = ascii_whitespace
    else:
        empty = ''
        default_replacement = ' '
        default_separators = whitespace

    if not s:
        return empty

    if separators is None:
        separators = default_separators
    if replacement is None:
        replacement = default_replacement

    if separators in (whitespace, ascii_whitespace, utf8_whitespace, None):
        if not s.strip():
            return replacement
        words = s.split()
        cleaned = replacement.join(words)
        del words
        if s[:1].isspace():
            cleaned = replacement + cleaned
        if s[-1:].isspace():
            cleaned = cleaned + replacement
        return cleaned

    words = list(multisplit(s, separators, keep=False, separate=False, strip=True, reverse=False, maxsplit=-1))
    cleaned = replacement.join(words)
    del words
    stripped_left = multistrip(s, separators, left=True, right=False) != s
    if stripped_left:
        cleaned = replacement + cleaned
    stripped_right = multistrip(s, separators, left=False, right=True) != s
    if stripped_right:
        cleaned = cleaned + replacement
    return cleaned


@_export
def split_quoted_strings(s, quotes=None, *, triple_quotes=True, backslash=None):
    """
    Splits s into quoted and unquoted segments.  Returns an iterator yielding
    2-tuples:
        (is_quoted, segment)
    where segment is a substring of s, and is_quoted is true if the segment is
    quoted.  Joining all the segments together recreates s.

    quotes is an iterable of quote separators.  Note that split_quoted_strings
    only supports quote *characters*, as in, each quote separator must be exactly
    one character long.

    If triple_quotes is true, supports "triple-quoted" strings like Python.

    If backslash is a character, this character will quoting characters inside
    a quoted string, like the backslash character inside strings in Python.
    """
    as_bytes = isinstance(s, bytes)
    if as_bytes:
        empty = b''
        if quotes is None:
            quotes=(b'"', b"'")
        if backslash is None:
            backslash = b'\\'
        i = (s[i:i+1] for i in range(len(s)))
    else:
        empty = ''
        if quotes is None:
            quotes=('"', "'")
        if backslash is None:
            backslash = '\\'
        i = s
    empty_join = empty.join

    i = PushbackIterator(i)
    in_quote = None
    in_triple_quote = False
    in_backslash = False
    text = []
    for c in i:
        if not in_quote:
            # encountered character while not in quoted string.
            if c not in quotes:
                # general case. append unquoted character to our unquoted string.
                text.append(c)
                continue
            if not triple_quotes:
                # triple quotes are off, encountered quote.
                # flush unquoted string, start quoted string.
                if text:
                    yield False, empty_join(text)
                    text.clear()
                text.append(c)
                in_quote = c
                continue
            # scan for triple quotes.
            c2 = next(i)
            if c2 != c:
                # only a single quote mark. flush unquoted string, start quoted string.
                i.push(c2)
                if text:
                    yield False, empty_join(text)
                    text.clear()
                text.append(c)
                in_quote = c
                continue
            c3 = i.next(None)
            if c3 != c:
                # two quotes in a row, but not three.
                # flush unquoted string, emit empty quoted string.
                if text:
                    yield False, empty_join(text)
                    text.clear()
                yield True, c*2
                if c3 is not None:
                    i.push(c3)
                continue
            # triple quoted string.
            # flush unquoted string, start triple quoted string.
            if text:
                yield False, empty_join(text)
                text.clear()
            text.append(c*3)
            in_quote = c
            in_triple_quote = c
            continue

        # handle quoted string
        if in_backslash:
            # previous character was a backslash.
            # append this character no matter what it is.
            text.append(c)
            in_backslash = False
            continue
        if c == backslash:
            # encountered backslash character.
            # set flag so we append the next character,
            # no matter what it is.
            in_backslash = True
            text.append(c)
            continue
        if c != in_quote:
            # character doesn't match our quote marker.
            # append to our quoted string.
            text.append(c)
            continue
        if not in_triple_quote:
            # we found our quote mark, and we're only
            # in single quotes (not triple quotes).
            # finish and emit the quoted string,
            # and return to unquoted mode.
            text.append(c)
            yield True, empty_join(text)
            text.clear()
            in_quote = False
            continue
        # we're in a triple-quoted string,
        # and found one of our quote marks.
        # scan to see if we got three in a row.
        c2 = i.next(None)
        c3 = i.next(None)
        if c == c2 == c3:
            # we found triple quotes.
            # finish and emit the triple-quoted string,
            # and return to unquoted mode.
            text.append(c*3)
            yield True, empty_join(text)
            text.clear()
            in_triple_quote = in_quote = False
            continue
        # didn't find triple quotes.  append the single
        # quote mark to our quoted string and push the
        # other two characters back onto the iterator.
        text.append(c)
        if c2 is not None:
            if c3 is not None:
                i.push(c3)
            i.push(c2)

    # flush the remainder of the string we were building,
    # in whatever condition it's in.
    if text:
        yield in_quote, empty_join(text)


@_export
class LineInfo:
    """
    The info object yielded by a lines iterator.
    You can add your own fields by passing them in
    via **kwargs; you can also add new attributes
    or modify existing attributes as needed from
    inside a "lines modifier" function.
    """
    def __init__(self, line, line_number, column_number, **kwargs):
        if not isinstance(line, (str, bytes)):
            raise TypeError("line must be str or bytes")
        if not isinstance(line_number, int):
            raise TypeError("line_number must be int")
        if not isinstance(column_number, int):
            raise TypeError("column_number must be int")
        self.line = line
        self.line_number = line_number
        self.column_number = column_number
        self.__dict__.update(kwargs)

    def __repr__(self):
        names = list(self.__dict__)
        priority_names = ['line', 'line_number', 'column_number']
        fields = []
        for name in priority_names:
            names.remove(name)
        names.sort()
        names = priority_names + names
        for name in names:
            fields.append(f"{name}={getattr(self, name)!r}")
        text = ", ".join(fields)
        return f"LineInfo({text})"

    def __eq__(self, other):
        return isinstance(other, self.__class__) and (other.__dict__ == self.__dict__)


@_export
class lines:
    def __init__(self, s, separators=None, *, line_number=1, column_number=1, tab_width=8, **kwargs):
        """
        A "lines iterator" object.  Splits s into lines, and iterates yielding those lines.

        "s" can be str, bytes, or any iterable.

        By default, if "s" is str, splits "s" by all Unicode line break characters.
        If "s" is bytes, splits "s" by all ASCII line break characters.

        If "s" is neither str nor bytes, "s" must be an iterable;
        lines yields successive elements of "s" as lines.

        "separators", if not None, must be an iterable of strings of the
        same type as "s".  lines will split "s" using those strings as
        separator strings (using big.multisplit).

        When iterated over, yields 2-tuples:
            (info, line)

        info is a LineInfo object, which contains three fields by default:
            * line - the original line, never modified
            * line_number - the line number of this line, starting at the
              line_number passed in and adding 1 for each successive line
            * column_number - the column this line starts on,
              starting at the column_number passed in, and adjusted when
              characters are removed from the beginning of line

        tab_width is not used by lines itself, but is stored internally and
        may be used by other lines modifier functions
        (e.g. lines_convert_tabs_to_spaces). Similarly, all keyword
        arguments passed in via kwargs are stored internally and can be
        accessed by user-defined lines modifier functions.

        Composable with all the lines_ functions from the big.text module.
        """
        if not isinstance(line_number, int):
            raise TypeError("line_number must be int")
        if not isinstance(column_number, int):
            raise TypeError("column_number must be int")
        if not isinstance(tab_width, int):
            raise TypeError("tab_width must be int")

        as_bytes = isinstance(s, bytes)
        as_str = isinstance(s, str)
        if as_bytes or as_str:
            if not separators:
                separators = newlines if as_str else ascii_newlines
            i = multisplit(s, separators, keep=False, separate=True, strip=False)
        else:
            i = iter(s)
            as_bytes = None

        self.s = s
        self.separators = separators
        self.line_number = line_number
        self.column_number = column_number
        self.tab_width = tab_width
        self.as_bytes = as_bytes

        self.i = i

        self.__dict__.update(kwargs)

    def __iter__(self):
        return self

    def __next__(self):
        line = next(self.i)
        if self.as_bytes is None:
            self.as_bytes = isinstance(line, bytes)
        return_value = (LineInfo(line, self.line_number, self.column_number), line)
        self.line_number += 1
        return return_value

@_export
def lines_rstrip(li):
    """
    A lines modifier function.  Strips trailing whitespace from the
    lines of a "lines iterator".

    Composable with all the lines_ functions
    from the big text module.
    """
    for info, line in li:
        yield (info, line.rstrip())

@_export
def lines_strip(li):
    """
    A lines modifier function.  Strips leading and trailing whitespace
    from the lines of a "lines iterator".

    If lines_strip removes leading whitespace from a line, it adds
    a field to the associated LineInfo object:
      * leading - the leading whitespace string that was removed

    Composable with all the lines_ functions from the big.text module.
    """
    for info, line in li:
        lstripped = line.lstrip()
        if not lstripped:
            line = lstripped
        else:
            original_leading = line[:len(line) - len(lstripped)]
            if original_leading:
                leading = original_leading.expandtabs(li.tab_width)
                info.column_number += len(leading)
                line = lstripped.rstrip()
                info.leading = original_leading

        yield (info, line)

@_export
def lines_filter_comment_lines(li, comment_separators):
    """
    A lines modifier function.  Filters out comment lines from the
    lines of a "lines iterator".  Comment lines are lines whose first
    non-whitespace characters appear in the iterable of
    comment_separators strings passed in.

    What's the difference between lines_strip_comments and
    lines_filter_comment_lines?
      * lines_filter_comment_lines only recognizes lines that
        *start* with a comment separator (ignoring leading
        whitespace).  Also, it filters out those lines
        completely, rather than modifying the line.
      * lines_strip_comments handles comment characters
        anywhere in the line, although it can ignore
        comments inside quoted strings.  It truncates the
        line but still always yields the line.

    Composable with all the lines_ functions from the big.text module.
    """
    if not comment_separators:
        raise ValueError("illegal comment_separators")

    if isinstance(comment_separators, bytes):
        comment_separators = tuple(comment_separators[i:i+1] for i in range(len(comment_separators)))
        as_bytes = True
    else:
        as_bytes = isinstance(comment_separators[0], bytes)

    comment_pattern = _separators_to_re(tuple(comment_separators), as_bytes, separate=False, keep=False)
    comment_re = re.compile(comment_pattern)
    for info, line in li:
        s = line.lstrip()
        if comment_re.match(s):
            continue
        yield (info, line)


@_export
def lines_containing(li, s, *, invert=False):
    """
    A lines modifier function.  Only yields lines
    that contain s.  (Filters out lines that
    don't contain s.)

    If invert is true, returns the opposite--
    filters out lines that contain s.

    Composable with all the lines_ functions from the big.text module.
    """
    if invert:
        for t in li:
            if not s in t[1]:
                yield t
        return
    for t in li:
        if s in t[1]:
            yield t

@_export
def lines_grep(li, pattern, *, invert=False, flags=0):
    """
    A lines modifier function.  Only yields lines
    that match the regular expression pattern.
    (Filters out lines that don't match pattern.)

    pattern can be str, bytes, or an re.Pattern object.
    If pattern is not an re.Pattern object, it's compiled
    with re.compile(pattern, flags=flags).

    If invert is true, returns the opposite--
    filters out lines that match pattern.

    Composable with all the lines_ functions from the big.text module.
    """
    if not isinstance(pattern, re.Pattern):
        pattern = re.compile(pattern, flags=flags)
    search = pattern.search
    if invert:
        for t in li:
            if not search(t[1]):
                yield t
        return
    for t in li:
        if search(t[1]):
            yield t

@_export
def lines_sort(li, *, reverse=False):
    """
    A lines modifier function.  Sorts all
    input lines before yielding them.

    Lines are sorted lexicographically,
    from lowest to highest.
    If reverse is true, lines are sorted
    from highest to lowest.

    Composable with all the lines_ functions from the big.text module.
    """
    lines = list(li)
    lines.sort(key=lambda t:t[1], reverse=reverse)
    yield from iter(lines)

@_export
def lines_strip_comments(li, comment_separators, *, quotes=('"', "'"), backslash='\\', rstrip=True, triple_quotes=True):
    """
    A lines modifier function.  Strips comments from the lines
    of a "lines iterator".  Comments are substrings that indicate
    the rest of the line should be ignored; lines_strip_comments
    truncates the line at the beginning of the leftmost comment
    separator.

    If rstrip is true (the default), lines_strip_comments calls
    the rstrip() method on line after it truncates the line.

    If quotes is true, it must be an iterable of quote characters.
    (Each quote character MUST be a single character.)
    lines_strip_comments will parse the line and ignore comment
    characters inside quoted strings.  If quotes is false,
    quote characters are ignored and line_strip_comments will
    truncate anywhere in the line.

    backslash and triple_quotes are passed in to
    split_quoted_string, which is used internally to detect
    the quoted strings in the line.

    Sets a new field on the associated LineInfo object for every line:
      * comment - the comment stripped from the line, if any.
        if no comment was found, "comment" will be an empty string.

    What's the difference between lines_strip_comments and
    lines_filter_comment_lines?
      * lines_filter_comment_lines only recognizes lines that
        *start* with a comment separator (ignoring leading
        whitespace).  Also, it filters out those lines
        completely, rather than modifying the line.
      * lines_strip_comments handles comment characters
        anywhere in the line, although it can ignore
        comments inside quoted strings.  It truncates the
        line but still always yields the line.

    Composable with all the lines_ functions from the big.text module.
    """
    if not comment_separators:
        raise ValueError("illegal comment_separators")

    if isinstance(comment_separators, bytes):
        comment_separators = tuple(comment_separators[i:i+1] for i in range(len(comment_separators)))
        as_bytes = True
    else:
        as_bytes = isinstance(comment_separators[0], bytes)

    if as_bytes:
        empty = b''
    else:
        empty = ''
    empty_join = empty.join

    comment_pattern = _separators_to_re(comment_separators, as_bytes=as_bytes, separate=True, keep=True)
    re_comment = re.compile(comment_pattern)
    split = re_comment.split
    for info, line in li:
        if quotes:
            i = split_quoted_strings(line, quotes, backslash=backslash, triple_quotes=triple_quotes)
        else:
            i = ((False, line),)
        segments = []
        append = segments.append
        comment = []
        for is_quoted, segment in i:
            if comment:
                comment.append(segment)
                continue
            if is_quoted:
                append(segment)
                continue
            fields = split(segment, maxsplit=1)
            leading = fields[0]
            if len(fields) == 1:
                append(leading)
                continue
            # found a comment marker in an unquoted segment!
            if rstrip:
                leading = leading.rstrip()
            append(leading)
            comment = fields[1:]

        info.comment = empty_join(comment)
        line = empty_join(segments)
        yield (info, line)

@_export
def lines_convert_tabs_to_spaces(li):
    """
    A lines modifier function.  Converts tabs to spaces for the lines
    of a "lines iterator", using the tab_width passed in to lines.

    Composable with all the lines_ functions from the big.text module.
    """
    for info, line in li:
        yield (info, line.expandtabs(li.tab_width))


@_export
def lines_strip_indent(li):
    """
    A lines modifier function.  Automatically measures and strips indents.

    Sets two new fields on the associated LineInfo object for every line:
      * indent - an integer indicating how many indents it's observed
      * leading - the leading whitespace string that was removed

    Uses an intentionally simple algorithm.
    Only understands tab and space characters as indent characters.
    Internally detabs to spaces first for consistency, using the
    tab_width passed in to lines.

    You can only dedent out to a previous indent.
    Raises IndentationError if there's an illegal dedent.

    Composable with all the lines_ functions from the big.text module.
    """
    indent = 0
    leadings = []
    for info, line in li:
        lstripped = line.lstrip()
        original_leading = line[:len(line) - len(lstripped)]
        leading = original_leading.expandtabs(li.tab_width)
        len_leading = len(leading)
        # print(f"{leadings=} {line=} {leading=} {len_leading=}")
        if leading.rstrip(' '):
            raise ValueError(f"lines_strip_indent can't handle leading whitespace character {leading[0]!r}")
        if not leading:
            indent = 0
            leadings.clear()
            new_indent = False
        elif not leadings:
            new_indent = True
        elif leadings[-1] == len_leading:
            new_indent = False
        elif len_leading > leadings[-1]:
            new_indent = True
        else:
            # not equal, not greater than... must be less than!
            new_indent = False
            assert leadings
            leadings.pop()
            indent -= 1
            while leadings:
                if leadings[-1] == len_leading:
                    break
                if leadings[-1] > len_leading:
                    new_indent = None
                    break
                leadings.pop()
                indent -= 1
            if not leadings:
                new_indent = None

        # print(f"  >> {leadings=} {new_indent=}")
        if new_indent:
            leadings.append(len_leading)
            indent += 1
        elif new_indent is None:
            raise IndentationError(f"line {info.line_number} column {len_leading + info.column_number}: unindent does not match any outer indentation level")
        info.leading = original_leading
        info.indent = indent
        if len_leading:
            info.column_number += len_leading
            line = lstripped

        yield (info, line)

@_export
def lines_filter_empty_lines(li):
    """
    A lines modifier function.  Filters out the empty lines
    of a "lines iterator".

    Preserves the line numbers.  If lines 0 through 2 are empty,
    line 3 is "a", line 4 is empty, and line 5 is "b",
    will yield:
        (3, "a")
        (5, "b")

    Composable with all the lines_ functions from the big.text module.
    """
    for t in li:
        if not t[1]:
            continue
        yield t



@_export
def wrap_words(words, margin=79, *, two_spaces=True):
    """
    Combines 'words' into lines and returns the result as a string.
    Similar to textwrap.wrap.

    'words' should be an iterator containing text split at word
    boundaries.  Example:
        "this is an example of text split at word boundaries".split()
    If you want a paragraph break, embed two '\n' characters in a row.

    'margin' specifies the maximum length of each line. The length of
    every line will be less than or equal to 'margin', unless the length
    of an individual element inside 'words' is greater than 'margin'.

    If 'two_spaces' is true, elements from 'words' that end in
    sentence-ending punctuation ('.', '?', and '!') will be followed
    by two spaces, not one.

    Elements in 'words' are not modified; any leading or trailing
    whitespace will be preserved.  You can use this to preserve
    whitespace where necessary, like in code examples.
    """
    words = iter(words)
    col = 0
    lastword = ''
    text = []

    for word in words:
        if word.isspace():
            lastword = word
            col = 0
            text.append(word)
            continue

        l = len(word)

        if two_spaces and lastword.endswith(('.', '?', '!')):
            space = "  "
            len_space = 2
        else:
            space = " "
            len_space = 1

        if (l + len_space + col) > margin:
            if col:
                text.append('\n')
                col = 0
        elif col:
            text.append(space)
            col += len_space

        text.append(word)
        col += len(word)
        lastword = word

    return "".join(text)



_code_paragraph = "code paragraph"
_text_paragraph = "text paragraph"

class _column_wrapper_splitter:

    def __init__(self, tab_width, allow_code, code_indent, convert_tabs_to_spaces):
        # print(f"\n_column_wrapper_splitter({tab_width=}, {allow_code=}, {convert_tabs_to_spaces=})")
        self.tab_width = tab_width
        self.allow_code = allow_code
        self.code_indent = code_indent
        self.convert_tabs_to_spaces = convert_tabs_to_spaces

        self.init()

    def init(self):
        self.words = []

        self.leading = []
        self.word = []
        self.code = []

        # used to regulate line and paragraph breaks.
        # can be _code_paragraph, _text_paragraph, or None.
        # (if None, we've already handled the appropriate break.)
        self.previous_paragraph = None

        self.state = self.state_initial

    def emit(self, c):
        # print(f" [emit]", repr(c))
        self.words.append(c)

    def line_break(self):
        # print(f" [  \\n]")
        self.words.append('\n')

    def paragraph_break(self):
        # print(f" [\\n\\n]")
        self.words.append('\n\n')

    def next(self, state, leading=None, word=None):
        # print(f" [  ->]", state.__name__, repr(leading), repr(word))
        self.state = state
        assert ((leading is None) and (word is None)) or ((leading is not None) and (word is not None))
        if word is not None:
            self.state(leading, word)

    def write(self, s):
        # write consumes s and makes calls as appropriate to
        # self.state().
        #
        # first, write aggregates together all consecutive
        # non-line-breaking whitespace characters, which it
        # stores in 'leading'.  if the next character is
        # a newline, it passes that single newline as 'word'.
        # otherwise it aggregates all consecutive non-whitespace
        # characters together, and passes those in as 'word'.
        #
        # thus we only call self.state() with two types of 'word':
        #     * a word, which is non-whitespace, and
        #     * a single '\n'.
        # all other whitespace is passed in as 'leading'.
        #
        # so for example, this input:
        #
        #   'hello there\nhow are you?\n\n  \n\ti am fine.\n  so there!'
        #
        # will result in self.state being called with these words,
        # and with whitespace being set to these values:
        #    leading   word
        #    -------   ------
        #    ''        'hello'
        #    ' '       'there'
        #    ''        '\n'
        #    ''        'how'
        #    ' '       'are'
        #    ' '       'you?'
        #    ''        '\n'
        #    ''        '\n'
        #    '  '      '\n'
        #    '\t'      'i'
        #    ' '       'am'
        #    ' '       'fine.'
        #    ''        '\n'
        #    '  '      'so'
        #    ' '       'there!'
        #
        # consecutive calls to write() behave the same as calling write()
        # once with both inputs concatenated together.  this:
        #     self.write('abc')
        #     self.write('def')
        # is the same as calling
        #     self.write('abcdef')
        #
        # you should call close() after the last write() call.

        leading = self.leading
        word = self.word

        for c in s:
            # print(f"<{c!r}> ", end='')

            if not c.isspace():
                word.append(c)
                continue

            if word:
                w = "".join(word)
                word.clear()
                if leading:
                    l = ''.join(leading)
                else:
                    l = ''
                leading.clear()
                self.state(l, w)
            if c == '\n':
                if leading:
                    l = ''.join(leading)
                else:
                    l = ''
                leading.clear()
                self.state(l, c)
            else:
                leading.append(c)

    def close(self):
        # flush the current word, if any.
        if self.word:
            self.state("".join(self.leading), "".join(self.word))
            self.leading.clear()
            self.word.clear()

    def state_paragraph_start(self, leading, word):
        """
        Initial state.  Also the state we return to
        after encountering two '\n's in a row after
        a text line.
        """
        if word == '\n':
            return
        if self.previous_paragraph:
            self.paragraph_break()
            self.previous_paragraph = None
        self.next(self.state_line_start, leading, word)

    state_initial = state_paragraph_start

    def state_line_start(self, leading, word):
        if word == '\n':
            # two '\n's in a row.
            if self.previous_paragraph == _code_paragraph:
                # we could still be in a code block.
                # remember the whitespace and continue.
                # we don't need to save the leading whitespace.
                self.code.append(word)
                return

            self.next(self.state_paragraph_start)
            return

        if self.allow_code:
            col = 0
            tab_width = self.tab_width
            for c in leading:
                if c == '\t':
                    col = col + tab_width - (col % tab_width)
                elif c == ' ':
                    col += 1
                else:
                    raise RuntimeError("unhandled whitespace character " + repr(c))
            if col >= self.code_indent:
                # code line!
                if self.previous_paragraph == _text_paragraph:
                    self.paragraph_break()
                    assert not self.code
                else:
                    if self.code:
                        for c in self.code:
                            self.line_break()
                        self.code.clear()

                self.next(self.state_code_line_start, leading, word)
                return

        if self.previous_paragraph == _code_paragraph:
            self.paragraph_break()
        self.next(self.state_text_line_start, leading, word)

    def state_text_line_start(self, leading, word):
        self.previous_paragraph = _text_paragraph
        self.next(self.state_in_text_line, leading, word)

    def state_in_text_line(self, leading, word):
        if word == '\n':
            self.next(self.state_line_start)
            return
        self.emit(word)

    def state_code_line_start(self, leading, word):
        self.previous_paragraph = _code_paragraph
        self.col = 0
        self.next(self.state_in_code_line, leading, word)

    def state_in_code_line(self, leading, word):
        if word == '\n':
            self.emit("".join(self.code))
            self.code.clear()
            self.next(self.state_line_start, '', word)
            return

        tab_width = self.tab_width
        convert_tabs_to_spaces = self.convert_tabs_to_spaces
        col = self.col
        for c in leading:
            if c == '\t':
                delta = tab_width - (col % tab_width)
                col += delta
                if convert_tabs_to_spaces:
                    self.code.append(' ' * delta)
                else:
                    self.code.append(c)
            elif c == ' ':
                col += 1
                self.code.append(c)
            else:
                raise RuntimeError("unhandled whitespace character " + repr(c))
        self.code.append(word)
        self.col = col + len(word)



@_export
def split_text_with_code(s, *, tab_width=8, allow_code=True, code_indent=4, convert_tabs_to_spaces=True):
    """
    Splits the string s into individual words,
    suitable for feeding into wrap_words.

    Paragraphs indented by less than code_indent will be
    broken up into individual words.

    If `allow_code` is true, paragraphs indented by at least
    code_indent spaces will preserve their whitespace:
    internal whitespace is preserved, and the newline is
    preserved.  (This will preserve the formatting of code
    examples, when these words are rejoined into lines
    by wrap_words.)
    """
    cws = _column_wrapper_splitter(tab_width, allow_code, code_indent, convert_tabs_to_spaces)
    for c in s:
        cws.write(c)
    cws.close()
    return cws.words


@_export
class OverflowStrategy(enum.Enum):
    """
    Enum providing constants to specify how merge_columns
    handles overflow in columns.
    """
    INVALID = enum.auto()
    RAISE = enum.auto()
    INTRUDE_ALL = enum.auto()
    # INTRUDE_MINIMUM = enum.auto()  # not implemented yet
    DELAY_ALL = enum.auto()
    # DELAY_MINIMUM = enum.auto()  # not implemented yet

@_export
def merge_columns(*columns, column_separator=" ",
    overflow_strategy=OverflowStrategy.RAISE,
    overflow_before=0,
    overflow_after=0,
    ):
    """
    Merge n column tuples, with each column tuple being
    formatted into its own column in the resulting string.
    Returns a string.

    columns should be an iterable of column tuples.
    Each column tuple should contain three items:
        (text, min_width, max_width)
    text should be a single text string, with newline
    characters separating lines. min_width and max_width
    are the minimum and maximum permissible widths for that
    column, not including the column separator (if any).

    Note that this function doesn't text-wrap the lines.

    column_separator is printed between every column.

    overflow_strategy tells merge_columns how to handle a column
    with one or more lines that are wider than that column's max_width.
    The supported values are:

        OverflowStrategy.RAISE

            Raise an OverflowError.  The default.

        OverflowStrategy.INTRUDE_ALL

           Intrude into all subsequent columns on all lines
           where the overflowed column is wider than its max_width.

        OverflowStrategy.DELAY_ALL

           Delay all columns after the overflowed column,
           not beginning any until after the last overflowed line
           in the overflowed column.

    When overflow_strategy is INTRUDE_ALL or DELAY_ALL, and
    either overflow_before or overflow_after is nonzero, these
    specify the number of extra lines before or after
    the overflowed lines in a column.
    """
    assert overflow_strategy in (OverflowStrategy.INTRUDE_ALL, OverflowStrategy.DELAY_ALL, OverflowStrategy.RAISE)
    raise_overflow_error = overflow_strategy == OverflowStrategy.RAISE
    delay_all = overflow_strategy == OverflowStrategy.DELAY_ALL

    _columns = columns
    columns = []
    empty_columns = []
    last_too_wide_lines = []
    max_lines = -1

    column_spacing = len(column_separator)

    for column_number, (s, min_width, max_width) in enumerate(_columns):

        # check types, let them raise exceptions as needed
        operator.index(min_width)
        operator.index(max_width)

        empty_columns.append(max_width * " ")

        if isinstance(s, str):
            lines = s.rstrip().split('\n')
        else:
            lines = s
        max_lines = max(max_lines, len(lines))

        # loop 1:
        # measure each line length, determining
        #  * maximum line length, and
        #  * all overflow lines
        rstripped_lines = []
        overflows = []
        max_line_length = -1
        in_overflow = False

        def add_overflow():
            nonlocal in_overflow
            in_overflow = False
            if overflows:
                last_overflow = overflows[-1]
                if last_overflow[1] >= (overflow_start - 1):
                    overflows.pop()
                    overflows.append((last_overflow[0], overflow_end))
                    return
            overflows.append((overflow_start, overflow_end))

        for line_number, line in enumerate(lines):
            line = line.rstrip()
            assert not "\n" in line
            rstripped_lines.append(line)

            length = len(line)
            max_line_length = max(max_line_length, length)

            line_overflowed = length > max_width
            if (not in_overflow) and line_overflowed:
                # starting new overflow
                if raise_overflow_error:
                    raise OverflowError(f"overflow in column {column_number}: {line!r} is {length} characters, column max_width is {max_width}")
                overflow_start = max(line_number - overflow_before, 0)
                in_overflow = True
            elif in_overflow and (not line_overflowed):
                # ending current overflow
                overflow_end = line_number - 1 + overflow_after
                add_overflow()

        if in_overflow:
            overflow_end = line_number + overflow_after
            add_overflow()
            for i in range(overflow_after):
                rstripped_lines.append('')

        if delay_all and overflows:
            overflows.clear()
            overflows.append((0, overflow_end))

        # loop 2:
        # compute padded lines and in_overflow for every line
        padded_lines = []
        overflows.reverse()
        overflow_start = overflow_end = None
        def next_overflow():
            nonlocal overflow_start
            nonlocal overflow_end
            if overflows:
                overflow_start, overflow_end = overflows.pop()
            else:
                overflow_start = overflow_end = sys.maxsize

        in_overflow = False
        next_overflow()
        for line_number, line in enumerate(lines):
            if line_number > overflow_end:
                in_overflow = False
                next_overflow()
            if line_number >= overflow_start:
                in_overflow = True
            if not in_overflow:
                line = line.ljust(max_width)
            padded_lines.append((line, in_overflow))

        columns.append(padded_lines)


    column_iterators = [iter(c) for c in columns]
    lines = []

    while True:
        line = []
        all_iterators_are_exhausted = True
        add_separator = False
        in_overflow = False
        for column_iterator, empty_column in zip(column_iterators, empty_columns):
            if add_separator:
                line.append(column_separator)
            else:
                add_separator = True

            try:
                column, in_overflow = next(column_iterator)
                all_iterators_are_exhausted = False
            except StopIteration:
                column = empty_column
            line.append(column)
            if in_overflow:
                break
        if all_iterators_are_exhausted:
            break
        line = "".join(line).rstrip()
        lines.append(line)

    text = "\n".join(lines)
    return text.rstrip()




#!/usr/bin/env python3

_license = """
big
Copyright 2022 Larry Hastings
All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

__all__ = ["Scheduler"]


from dataclasses import dataclass
import threading
from .heap import Heap
import time



invalid_event = object()

class Scheduler:
    """
    A replacement for Python's sched.scheduler object,
    adding full threading support and a modern Python interface.

    Python's sched.scheduler object is unfixable; its interface
    means it cannot correctly handle some scenarios:
        * If one thread has called sched.scheduler.run,
          and the next scheduled event will occur at time T,
          and a second thread schedules a new event which
          occurs at a time < T, sched.scheduler.run won't
          return any events to the first thread until time T.
        * If one thread has called sched.scheduler.run,
          and the next scheduled event will occur at time T,
          and a second thread cancels all events,
          sched.scheduler.run won't exit until time T.

    Also, sched.scheduler is thirty years behind the times in
    Python API design.  Its events are callbacks, which it
    calls.  Scheduler fixes this: its events are objects,
    and you iterate over the Scheduler object to receive
    events as they become due.

    Scheduler also benefits from thirty years of improvements
    to sched.scheduler, and in particular reimplements the
    entire (relevant portion of the) sched.scheduler test suite.
    """

    DEFAULT_PRIORITY = 100

    @dataclass(order=True)
    class Event:
        time: object
        priority: object
        sequence: int
        event: object
        scheduler: object

        def cancel(self):
            self.scheduler.cancel(self)

    class Regulator:
        """
        Regulator encapsulates all the Scheduler's needs
        for interacting with time.
        The Scheduler has no knowledge about time itself;
        it uses a Regulator object for all its time-related
        needs.

        You can implement your own Regulator and use it
        with Scheduler; the only methods it needs are
        now, sleep, and wake.
        """

        def __init__(self, scheduler):
            self.event = threading.Event()

        def __repr__(self): # pragma: no cover
            return f"<Regulator {event}>"

        def now(self):
            """
            Returns the current time in local units.
            Must be monotonically increasing; for any
            two calls to now during the course of the
            program, the later call must *never*
            have a lower value than the earlier call.

            (The same Scheduler object will never call
            now twice simulaneously on the same Regulator
            object, as all access to the Regulator object
            is done under a lock.)
            """
            return time.time()

        def wake(self):
            """
            Aborts all current calls to sleep,
            across all threads.
            """
            self.event.set()
            self.event.clear()

        def sleep(self, interval):
            """
            Sleeps for some amount of time, in local units.
            Must support an interval of 0.
            It's preferable that an interval of 0 yields the rest
            of the current thread's current time slice.

            If wake is called on this Regulator object while a
            different thread has called this function to sleep,
            sleep must abandon the rest of the sleep interval
            and return immediately.
            """
            self.event.wait(interval)

    def __init__(self, regulator=None):
        if regulator is None:
            regulator = self.Regulator(self)
        self.regulator = regulator
        self.heap = Heap()
        self.lock = threading.Lock()
        self.counter = 0

    def schedule(self, o, time, *, absolute=False, priority=DEFAULT_PRIORITY):
        """Schedule a new event to be yielded at a specific future time.

        By default, "time" is relative to the current time.
        If absolute is true, "time" is an absolute time.

        "priority" represents the importance of the event.
        Lower numbers are more important.
        There are no predefined priorities; you may use it however you like.

        Returns an object which can be used to remove the event from the queue:
           o = scheduler.schedule(my_event, 5)
           o.cancel() # cancels the event
        Alternately you can pass in this object to the scheduler's cancel method:
           o = scheduler.schedule(my_event, 5)
           scheduler.cancel(o)
        """
        if o == invalid_event:
            raise ValueError('invalid event')
        if not absolute:
            time += self.regulator.now()
        with self.lock:
            self.counter += 1
            event = self.Event(time, priority, self.counter, o, self)
            self.heap.append(event)
            self.regulator.wake()
        return event

    def cancel(self, event):
        """
        Cancel a scheduled event.

        event must be a value returned by add().
        If the event is not in the queue, raises ValueError.
        """
        with self.lock:
            self.heap.remove(event)
            self.regulator.wake()

    def __bool__(self):
        """
        Returns True if the queue is not empty.
        """
        with self.lock:
            return bool(self.heap)

    @property
    def queue(self):
        """
        An ordered list of upcoming events.

        The entries in the list are big.scheduler.Event objects.
        """
        with self.lock:
            return self.heap.queue

    def _next(self, blocking=True):
        # why loop here? in case we get awakened by the queue changing.
        #
        # also, I swear that back in my early Windows days, sleeping on
        # timers wasn't perfect and would often return slightly early.
        while True:
            # print(f"{self._regulator.now()}  {self.heap=}")
            # don't sleep while holding the lock!
            # we write down what we need to do in
            # local variables, then do it after releasing
            # the lock.
            sleep_interval = 0
            event = invalid_event

            with self.lock:
                if not self.heap:
                    # print(" empty queue DONE")
                    raise StopIteration

                sleep = self.regulator.sleep

                ev = self.heap[0]

                now = self.regulator.now()
                interval = ev.time - now
                ready = interval <= 0

                if not ready:
                    if not blocking:
                        # print(" wouldblock DONE")
                        raise StopIteration
                    sleep_interval = max(interval, 0)
                else:
                    self.heap.popleft()
                    event = ev.event

            # we always delay here.
            # if sleep_interval is 0,
            # this still yields the remainder of our timeslice
            # and lets other threads run.
            # print(f"  sleeping {sleep_interval}")
            sleep(sleep_interval)
            if event is not invalid_event:
                # print(f"  returning {event}")
                return event

    def __iter__(self):
        return self

    def __next__(self):
        return self._next()


    class NonBlockingSchedulerIterator:
        def __init__(self, scheduler):
            self.s = scheduler

        def __iter__(self):
            return self

        def __next__(self):
            return self.s._next(blocking=False)

    def non_blocking(self):
        """
        A non-blocking iterator over the queue.
        Only yields events that are currently ready,
        then stops iteration.
        """
        return self.NonBlockingSchedulerIterator(self)




#!/usr/bin/env python3

_license = """
big
Copyright 2022 Larry Hastings
All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import builtins
import os
import pathlib
import re


__all__ = []

def export(o):
    __all__.append(o.__name__)
    return o


import shutil
which = export(shutil.which)


@export
def fgrep(path, text, *, encoding=None, enumerate=False, case_insensitive=False):
    """
    Find the lines of a file that match some text, like the UNIX "fgrep"
    utility program.

    path should be an object representing a path to an existing file, one of:
      * a string,
      * a bytes object, or
      * a pathlib.Path object.

    text should be either string or bytes.

    encoding is used as the file encoding when opening the file.

    if text is a str, the file is opened in text mode.
    if text is a bytes object, the file is opened in binary mode.
    encoding must be None when the file is opened in binary mode.

    If case_insensitive is true, perform the search in a case-insensitive
    manner.

    Returns a list of lines in the file containing text.  The lines are either
    strings or bytes objects, depending on the type of text.  The lines
    have their newlines stripped but preserve all other whitespace.

    If enumerate is true, returns a list of tuples of (line_number, line).
    The first line of the file is line number 1.

    For simplicity of implementation, the entire file is read in to memory
    at one time.  If `case_insensitive` is True, a lowercased copy is also used.
    """
    if isinstance(text, bytes):
        if encoding is not None:
            raise ValueError("encoding must be None when text is bytes")
        mode = 'rb'
        separator = b'\n'
    else:
        mode = 'rt'
        separator = '\n'
    if isinstance(path, pathlib.Path):
        f = path.open(mode, encoding=encoding)
    else:
        f = open(path, mode, encoding=encoding)
    with f:
        contents = f.read()
        split = contents.split(separator)
        if not case_insensitive:
            if enumerate:
                return [t for t in builtins.enumerate(split, 1) if text in t[1]]
            return [line for line in split if text in line]

        lower_contents = contents.lower()
        lower_split = lower_contents.split(separator)
        text = text.lower()
        if enumerate:
            return [(line_number, line) for line_number, (line, lower_line) in builtins.enumerate(zip(split, lower_split), 1) if text in lower_line]
        return [line for line, lower_line in zip(split, lower_split) if text in lower_line]


@export
def grep(path, pattern, *, encoding=None, enumerate=False, flags=0):
    """
    Look for matches to a regular expression pattern in the lines of a file,
    like the UNIX "grep" utility program.

    path should be an object representing a path to an existing file, one of:
      * a string,
      * a bytes object, or
      * a pathlib.Path object.

    pattern should be an object containing a regular expression, one of:
      * a string,
      * a bytes object, or
      * an re.Pattern, initialized with either str or bytes.

    encoding is used as the file encoding when opening the file.

    if pattern uses a str, the file is opened in text mode.
    if pattern uses a bytes object, the file is opened in binary mode.
    encoding must be None when the file is opened in binary mode.

    flags is passed in as the flags argument to re.compile if pattern
    is a string.

    Returns a list of lines in the file matching the pattern.  The lines
    are either strings or bytes objects, depending on the type of pattern.
    The lines have their newlines stripped but preserve all other whitespace.

    If enumerate is true, returns a list of tuples of (line_number, line).
    The first line of the file is line number 1.

    For simplicity of implementation, the entire file is read in to memory
    at one time.

    (Tip: to perform a case-insensitive pattern match, pass in the
    re.IGNORECASE flag into flags for this function (if pattern is a string
    or bytes) or when creating your regular expression object (if pattern is
    an re.Pattern object).
    """
    if not isinstance(pattern, re.Pattern):
        pattern = re.compile(pattern, flags)
    if isinstance(pattern.pattern, bytes):
        if encoding is not None:
            raise ValueError("encoding must be None when pattern uses bytes")
        mode = 'rb'
        separator = b'\n'
    else:
        mode = 'rt'
        separator = '\n'
    if isinstance(path, pathlib.Path):
        f = path.open(mode, encoding=encoding)
    else:
        f = open(path, mode, encoding=encoding)
    with f:
        text = f.read()
        split = text.split(separator)
        if enumerate:
            return [t for t in builtins.enumerate(split, 1)
                    if pattern.search(t[1])]
        return [line for line in split if pattern.search(line)]


@export
class pushd:
    """
    A context manager that temporarily changes the directory.
    Example:

    with big.pushd('x'):
        pass

    This would change into the `'x'` subdirectory before
    executing the nested block, then change back to
    the original directory after the nested block.

    You can change directories in the nested block;
    this won't affect pushd restoring the original current
    working directory upon exiting the nested block.
    """
    def __init__(self, path):
        self.path = path
        self.cwd = os.getcwd()
    def __enter__(self):
        os.chdir(self.path)
    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self.cwd)


@export
def safe_mkdir(path):
    """
    Ensures that a directory exists at 'path'.
    If this function returns and doesn't raise,
    it guarantees that a directory exists at 'path'.

    If a directory already exists at 'path',
    does nothing.

    If a file exists at 'path', unlinks it
    then creates the directory.

    If the parent directory doesn't exist,
    creates it, then creates 'path'.

    This function can still fail:
      * 'path' could be on a read-only filesystem.
      * You might lack the permissions to create 'path'.
      * You could ask to create the directory 'x/y'
        and 'x' is a file (not a directory).
    """
    if os.path.isfile(path):
        os.unlink(path)
    os.makedirs(path, exist_ok=True)

@export
def safe_unlink(path):
    """
    Unlinks path, if path exists and is a file.
    """
    if os.path.isfile(path):
        os.unlink(path)

@export
def file_size(path):
    """
    Returns the size of the file at `path`, as an integer
    representing the number of bytes.
    """
    st = os.stat(path)
    return st.st_size

@export
def file_mtime(path):
    """
    Returns the modification time of path, in seconds since the epoch.
    Note that seconds is a float, indicating the sub-second with some
    precision.
    """
    st = os.stat(path)
    return st.st_mtime

@export
def file_mtime_ns(path):
    """
    Returns the modification time of path, in nanoseconds since the epoch.
    """
    st = os.stat(path)
    return st.st_mtime_ns

@export
def touch(path):
    """
    Ensures that path exists, and its modification time is the current time.

    If path does not exist, creates an empty file.

    If path exists, updates its modification time to the current time.
    """
    if os.path.exists(path):
        os.utime(path)
    else:
        with open(path, "wb") as f:
            pass

# exfat allowed characters:
#
# all Unicode characters except
#     U+0000 (NUL) through U+001F (US)
#     / (slash)
#     \ (backslash)
#     : (colon)
#     * (asterisk)
#     ? (question mark)
#     " (quote)
#     < (less than)
#     > (greater than)
# and | (pipe)
#
# https://en.wikipedia.org/wiki/ExFAT
_exfat_translation_dict = {
    "\x00": "@",
    "\x01": "?01",
    "\x02": "?02",
    "\x03": "?03",
    "\x04": "?04",
    "\x05": "?05",
    "\x06": "?06",
    "\x07": "?07",
    "\x08": "?08",
    "\x09": "?09",
    "\x0a": "?0a",
    "\x0b": "?0b",
    "\x0c": "?0c",
    "\x0d": "?0d",
    "\x0e": "?0e",
    "\x0f": "?0f",

    "\x10": "?10",
    "\x11": "?11",
    "\x12": "?12",
    "\x13": "?13",
    "\x14": "?14",
    "\x15": "?15",
    "\x16": "?16",
    "\x17": "?17",
    "\x18": "?18",
    "\x19": "?19",
    "\x1a": "?1a",
    "\x1b": "?1b",
    "\x1c": "?1c",
    "\x1d": "?1d",
    "\x1e": "?1e",
    "\x1f": "?1f",

    "/": "-",
    "\\": "-",
    "*": "@",
    "?": ".",
    '"': "'",
    '<': "{",
    '>': "}",
    '|': "!",
    }

_exfat_translation_table = ''.maketrans(_exfat_translation_dict)

@export
def translate_filename_to_exfat(s):
    """
    Ensures that all characters in s are legal for a FAT filesystem.

    Returns a copy of s where every character not allowed in a FAT
    filesystem filename has been replaced with a character (or characters)
    that are permitted.
    """
    if not (isinstance(s, str) and s):
        raise ValueError("filename can't be empty")
    s = s.translate(_exfat_translation_table)
    s = s.replace(": ", " - ")
    s = s.replace(":", "-")
    return s


# unix filesystem allowed characters:
# everything except a zero byte and '/'
_unix_translation_dict = {
        "\x00": "@",
        "/":    "-",
    }

_unix_translation_table = ''.maketrans(_unix_translation_dict)

@export
def translate_filename_to_unix(s):
    """
    Ensures that all characters in s are legal for a UNIX filesystem.

    Returns a copy of s where every character not allowed in a UNIX
    filesystem filename has been replaced with a character (or characters)
    that are permitted.
    """
    if not (isinstance(s, str) and s):
        raise ValueError("filename can't be empty")
    return s.translate(_unix_translation_table)



#!/usr/bin/env python3

_license = """
big
Copyright 2022 Larry Hastings
All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


#
# TODO
#
# * change _default_view to lazily initialize from _stock_view?
#

from collections import defaultdict

try:
    from graphlib import CycleError
except ImportError: # pragma: no cover
    class CycleError(ValueError):
        """
        Exception thrown by TopologicalSorter when it detects a cycle.
        """
        pass

__all__ = ["TopologicalSorter", "CycleError"]



nodes_dict_default = lambda: [0, {}]

class TopologicalSorter:
    """
    An object representing a directed graph of nodes.

    See Python's graphlib.TopologicalSorter for concepts
    and the basic API.

    https://docs.python.org/3/library/graphlib.html#graphlib.TopologicalSorter
    """

    class View:
        """
        A view on a TopologicalSorter graph object.
        Allows iterating over the nodes of the graph
        in dependency order.
        """
        def __init__(self, graph, *, original=None):
            if not isinstance(graph, TopologicalSorter):
                raise ValueError("invalid graph")

            self.graph = graph
            self.graph.views.append(self)

            self._conflicts = defaultdict(set)
            self._reverse_conflicts = defaultdict(set)

            if original is not None:
                self._predecessors = original._predecessors.copy()
                self._ready = original._ready.copy()
                self._yielded = original._yielded.copy()
                self._done = original._done.copy()

                for n, v in original._conflicts.items():
                    self._conflicts[n] = v.copy()
                for n, v in original._reverse_conflicts.items():
                    self._reverse_conflicts[n] = v.copy()
                return

            # self._predecessors maps nodes to their
            # predecessor count in the local view.
            # we use this to discover ready nodes.
            self._predecessors = {}

            # self._ready is the set of currently-ready nodes.
            # they haven't been returned by ready() yet.
            self._ready = []

            # _yielded and _done are sets, because
            # they're definitely faster as sets,
            # and their ordering (or lack thereof)
            # doesn't affect the behavior of the view.

            # self._yielded is the set of nodes
            # yielded to the user by ready().
            self._yielded = set()

            # self._done is the set of nodes
            # returned from the user by done().
            self._done = set()

        def __repr__(self):
            return f"<View {hex(id(self))} on graph={hex(id(self.graph))} {self._ready=} yielded={len(self._yielded)} done={len(self._done)} conflicts={bool(self._conflicts)}>"

        def copy(self):
            """
            Returns a shallow copy of the view, duplicating its current state.
            """
            return self.__class__(self.graph, original=self)

        def reset(self):
            """
            Resets the view to its initial state,
            forgetting all "ready" and "done" state.
            """
            if self.graph is None:
                raise ValueError("TopologicalSorter view is already closed")
            self._predecessors.clear()
            self._predecessors.update(self.graph._stock_view._predecessors)
            self._ready.clear()
            self._ready.extend(self.graph._stock_view._ready)
            self._yielded.clear()
            self._done.clear()
            self._conflicts.clear()
            self._reverse_conflicts.clear()

        def close(self):
            """
            Closes the view.  A closed view can no longer be used.
            """
            if self.graph is None:
                raise ValueError("TopologicalSorter view is already closed")

            self.graph.views.remove(self)
            self.graph = None

            del self._predecessors
            del self._ready
            del self._yielded
            del self._done
            del self._conflicts
            del self._reverse_conflicts

        def __bool__(self):
            """
            Returns `True` if more work can be done in the
            view--if there are nodes waiting to be yielded by
            `get_ready`, or waiting to be returned by `done`.
            """
            if self.graph is None:
                raise ValueError("TopologicalSorter view has already been closed")
            if self._conflicts:
                raise RuntimeError("TopologicalSorter view is incoherent to the graph")
            return len(self._done) < len(self.graph.nodes)

        def ready(self):
            """
            Returns an iterable of all the nodes that
            currently have no dependencies.
            """
            if self.graph is None:
                raise ValueError("TopologicalSorter view has already been closed")
            if self._conflicts:
                raise RuntimeError("TopologicalSorter view is incoherent to the graph")
            if self.graph.dirty:
                self.graph.check_for_cycles()
            self._yielded.update(self._ready)
            ready = list(self._ready)
            self._ready.clear()
            return tuple(ready)

        def done(self, *nodes):
            """
            Marks nodes returned by ready as "done",
            possibly allowing additional nodes to be available
            from ready.
            """
            if self.graph is None:
                raise ValueError("TopologicalSorter view has already been closed")
            if self._conflicts:
                raise RuntimeError("TopologicalSorter view is incoherent to the graph")
            for node in nodes:
                if node not in self.graph.nodes:
                    raise ValueError(f"node {node!r} was not added using add()")
                if node not in self._yielded:
                    raise ValueError(f"node {node!r} was not passed out")
                # print("CALL _mark_node_done", node, self.graph.nodes[node])
                self._mark_node_done(node, self.graph.nodes[node][1])
            nodes = set(nodes)
            self._done |= nodes
            self._yielded -= nodes

        #
        # These functions are internal only.
        # They're a private API between the graph
        # and the views on the graph, or just internal
        # to the view itself.
        #
        # They're not public APIs, and it's not
        # supported for anyone but the graph to call them.
        #
        # In particular, they are optimized for exactly what
        # the graph needs to communicate to the views.
        # They aren't good general-purpose APIs and shouldn't
        # be used by any code outside this file.
        #
        # That enough warnings for you?  Don't touch!  I mean it!
        #
        def _add_nodes(self, nodes):
            """
            event: nodes have been added to the graph.
            "nodes" is an iterable of newly added nodes.

            Called *after* the nodes are initially added to the graph.
            All these nodes have already been added to the graph,
            and already have all their initial successors.

            We call _add_successor after this to notify about new successors.
            """
            for node in nodes:
                self._predecessors[node] = 0
                self._ready.append(node)

        def _add_successor(self, node, successor):
            """
            Called *after* the graph is updated.
            successor is guaranteed to be a new successor to node.
            """

            # CAPITAL LETTER is the predecessor node
            # single digit is the successor node
            # lowercase letter is the result
            #
            # so, A-1, 1 depends on A
            #
            # ---------
            # A  1  not yet yielded by ready (not ready, or waiting to be yielded by ready)
            # B  2  yielded by ready but not yet marked done
            # C  3  done
            # ---------
            #
            #   1__2__3
            # A|o* x  x
            # B|o* x  x
            # C|o* o  o
            #
            # o = ok!
            # x = not ok, view is incoherent.
            # * = if 1 was previously waiting in "ready", it's removed.

            conflicted = not (
                (node in self._done)
                or ((successor not in self._done)
                    and (successor not in self._yielded)) )
            if conflicted:
                # print("CONFLICT", node, successor, self._done)
                self._conflicts[node].add(successor)
                self._reverse_conflicts[successor].add(node)

            if successor not in self._done:
                self._predecessors[successor] += 1
                if self._predecessors[successor] == 1:
                    try:
                        # why "try" here?
                        # successor might be yielded or done.
                        # (in which case we must be conflicted.)
                        self._ready.remove(successor)
                    except ValueError:
                        pass
        def _remove_node(self, node, successors):
            """
            Called *after* the node is removed from the graph.
            The node is guaranteed to have been added to the view,
            either with _add_nodes, or was brought over by copy().
            """
            # print(f"{self=} _remove_node {node=} {successors=}")
            if (node in self._conflicts) or (node in self._reverse_conflicts):
                # print("REMOVE vYHS", node)
                for this, other in (
                    (self._conflicts, self._reverse_conflicts),
                    (self._reverse_conflicts, self._conflicts)
                    ):
                    values = this.get(node, ())
                    if values:
                        del this[node]
                        for v in values:
                            other[v].discard(node)
                            if not other[v]:
                                del other[v]

            if node in self._done:
                self._done.remove(node)
                return

            if node in self._ready:
                self._ready.remove(node)
            self._yielded.discard(node)

            self._predecessors.pop(node)

            self._mark_node_done(node, successors)

        def _mark_node_done(self, node, successors):
            """
            Does *not* update _done or _yielded.
            """
            # print(f"{self=} _mark_node_done {node=} {successors=}")

            # this "for" statement is why successors is a dict, not a set.
            # when iterating over a dict, you see the elements in insertion order.
            # when iterating over a set, you see the elements in random order.
            # by using a dict, the object's behavior is more stable,
            # which can definitely make debugging easier.
            for successor in successors:
                if (successor in self._yielded) or (successor in self._done):
                    continue
                self._predecessors[successor] -= 1
                if not self._predecessors[successor]:
                    self._ready.append(successor)
            # print(f"_mark_node_done done. {self=}")

        def print(self, print=print):
            """
            Prints the internal state of the view, and its graph.
            Used for debugging.

            print is the function used for printing;
            it should behave identically to the builtin print function.
            """
            print(self)
            print("      _ready:", sorted(self._ready))
            print("    _yielded:", sorted(self._yielded))
            print("       _done:", sorted(self._done))
            print("  _conflicts:", dict(self._conflicts))
            self.graph.print(print=print)
            print()

    def __bool__(self):
        """
        Returns True if more work can be done in the
        default view--if there are nodes waiting to be yielded
        by get_ready, or waiting to be returned by done.

        Aliased to is_active for compatibility
        with graphlib.
        """
        return bool(self._default_view)

    def __len__(self):
        """
        Returns the number of nodes in the graph.
        """
        return len(self.nodes)

    def ready(self):
        """
        Returns a tuple of "ready" nodes--nodes with no
        predecessors, or nodes whose predecessors have all
        been marked "done".

        Aliased to get_ready for compatibility with graphlib.
        """
        return self._default_view.ready()

    def done(self, *nodes):
        """
        Marks nodes returned by ready as "done",
        possibly allowing additional nodes to be available
        from ready.
        """
        return self._default_view.done(*nodes)

    def reset(self):
        """
        Resets the default view to its initial state,
        forgetting all "ready" and "done" state.
        """
        return self._default_view.reset()

    def __init__(self, graph=None):
        # the graph argument is a mapping of nodes to dependencies.

        # self.nodes maps node -> [count-of-predecessors, list-of-successors].
        # the first element of the list is the number of nodes *we* depend on.
        # the second element of the list is a set of nodes that depend on *us*.
        #
        # nodes added to the graph are always listed in self.nodes, even
        # if they've been given to the user by ready() or returned by done().
        self.nodes = {}

        # If the "dirty" attribute is true, we need to check for cycles.
        #
        # we're slightly smart about this:
        # we only set the dirty flag when we add an edge between
        # two nodes that were already in the graph.
        self.dirty = False

        self.views = []
        self._default_view = self.View(self)
        self._stock_view = self.View(self)

        if graph is not None:
            for node, dependencies in graph.items():
                self.add(node, *dependencies)

    def __repr__(self):
        return f"<TopologicalSorter {hex(id(self))} {len(self.nodes)} nodes>"

    def view(self):
        """
        Returns a new View object on this graph.
        """
        return self._stock_view.copy()

    def add(self, node, *dependencies):
        """
        Add a node to the graph if it isn't already there.
        All arguments are nodes of the graph; the first is
        the "node" added, and it will depend on all subsequent
        arguments.

        If add() is called multiple times with the same node
        argument, the set of dependencies will be the union of
        all dependencies passed in.

        It's legal to add a node with no dependencies.

        It's legal to depend on a node the graph hasn't otherwise
        seen yet.  It will be automatically added (and will have no
        dependencies).

        You can only add a dependency once;
        any further attempts are ignored.
        Calling K.add('a', 'b') twice is the same as
        calling K.add('a', 'b') once, or
        calling K.add('a', 'b', 'b', 'b', 'b') once.
        """
        # print(f"{node:3} depends on {dependencies!r}")

        new_nodes = []
        new_successors = []

        if not node in self.nodes:
            new_nodes.append(node)
            self.nodes[node] = nodes_dict_default()

        pc_s = self.nodes[node]
        predecessors_count, successors = pc_s

        for d in dependencies:
            if d not in self.nodes:
                new_nodes.append(d)
                self.nodes[d] = nodes_dict_default()
            d_pc, d_s = self.nodes[d]
            if node not in d_s:
                self.dirty = True
                d_s[node] = None
                predecessors_count += 1
                new_successors.append(d)

        pc_s[0] = predecessors_count

        if new_nodes:
            for v in self.views:
                v._add_nodes(new_nodes)
        if new_successors:
            for v in self.views:
                for s in new_successors:
                    v._add_successor(s, node)

    def remove(self, node):
        """
        Remove node from the graph.

        If any node P depends on a node N, this dependency
        is also removed, but P is not removed from the graph.

        remove() works but it's slow (O(N)).
        TopologicalSorter is optimized for fast adds and fast views.
        """

        # print("remove!", node)
        # print(f"    before {self.nodes=}")
        if node not in self.nodes:
            raise ValueError("node not in graph")

        predecessors = []
        predecessor_count, successors = self.nodes[node]
        if predecessor_count:
            # we have to iterate over all nodes in the graph
            # until we find all predecessors.
            for n, pc_s in self.nodes.items():
                pc, s = pc_s
                if node in s:
                    predecessor_count -= 1
                    predecessors.append(n)
                    # early exit! on average, O(N * 0.5) ;-)
                    # phew!
                    if not predecessor_count:
                        break

        # print(f"    {predecessors=}")
        # print(f"    {successors=}")

        for n in predecessors:
            _, s = self.nodes[n]
            del s[node]

        for n in successors:
            p_s = self.nodes[n]
            p_s[0] -= 1

        del self.nodes[node]
        # print(f"    after {self.nodes=}")
        # print()

        for v in self.views:
            v._remove_node(node, successors)

        # we never need to set the dirty flag here.
        # dirty is only used for detecting cycles.
        # you can't add a cycle by removing nodes or edges.

    def copy(self):
        """
        Returns a shallow copy of the graph.  The copy also
        duplicates the state of get_ready and done.
        """
        clone = self.__class__()
        for node, p_s in self.nodes.items():
            p, s = p_s
            clone.nodes[node] = [p, s.copy()]

        clone._default_view = self._default_view.copy()
        clone._default_view.graph = clone
        clone._stock_view = self._stock_view.copy()
        clone._stock_view.graph = clone

        return clone

    def cycle(self):
        """
        Cycle detector.
        If the graph has no cycles, returns None.
        If the graph has a cycle, returns a tuple containing
          nodes in a cycle.
        """
        if not self.dirty:
            # dirty bit is only cleared when we don't have a cycle.
            # ergo, if dirty bit isn't set, we have no cycles.
            return None
        self.dirty = False

        counts = {node: value[0] for node, value in self.nodes.items()}
        done = set()
        _ready = {node for node, count in counts.items() if not count}
        while _ready:
            done |= _ready
            _ready2 = set()
            for node in _ready:
                successors = self.nodes[node][1]
                for successor in successors:
                    counts[successor] -= 1
                    if not counts[successor]:
                        _ready2.add(successor)
            _ready = _ready2

        if len(done) == len(counts):
            # no cycles!
            return None

        # cycle detected.
        # start with all nodes that are aren't done yet.
        nodes = set(self.nodes) - done

        # pick a node at random and see if we can find a cycle.
        # note that this node may not participate in a cycle!
        # e:g:
        #   A -> B
        #   B -> A
        #   B -> C
        # C is not in a cycle, but it isn't done yet either.

        # the classic algorithm here is Depth First Search.
        # DFS is most clearly written recursively,
        # but this is Python and we could blow our stack.
        iterators = [iter(nodes)]
        seen = set()
        cycle = []

        while iterators:
            top = iterators[-1]
            try:
                node = next(top)
                # it's impossible for 'node' to be marked done.
                # we started with the list of nodes that weren't
                # marked 'done' yet, and the only thing we ever
                # add to 'iterators' are successor nodes.
                assert node not in done
                if node in seen:
                    # found a cycle
                    index = cycle.index(node)
                    self.dirty = True
                    return tuple(cycle[index:])
                # descend
                seen.add(node)
                cycle.append(node)
                iterators.append(iter(self.nodes[node][1]))
            except StopIteration:
                # ascend
                node = cycle.pop()
                seen.remove(node)
                iterators.pop()

    def check_for_cycles(self):
        """
        Checks for cycles.  If any cycles exist, raises CycleError.

        Aliased to prepare for compatibility with graphlib.
        """
        if self.dirty:
            cycle = self.cycle()
            if cycle:
                raise CycleError(f"nodes are in a cycle", cycle)

    def static_order(self):
        """
        Returns an iterator object that yields the
        graph's nodes in topological order.
        """
        v = self.view()
        while v:
            ready = v.ready()
            yield from ready
            v.done(*ready)

    def print(self, print=print):
        """
        Prints the internal state of the graph.  Used for debugging.

        print is the function used for printing;
        it should behave identically to the builtin print function.
        """
        print(self)
        print(f"  {len(self.nodes)} nodes:".rjust(13))
        for node, p_s in sorted(self.nodes.items()):
            print("         ", node, p_s)

    # graphlib 1.0 compatibility
    is_active = __bool__
    prepare = check_for_cycles
    get_ready = ready



if __name__ == "__main__": # pragma: no cover
    #
    # A simple "tsort" utility.
    #     https://en.wikipedia.org/wiki/Tsort
    #
    # reads stdin (or a file) containing pairs
    # of strings, e.g.:
    #    A B
    # That line would indicate that B comes after A
    # (B depends on A).

    import sys

    file = sys.stdin
    if len(sys.argv) > 1:
        filename = sys.argv[1]
        if filename != "-":
            file = open(filename, "rt")

    g = TopologicalSorter()
    for line in file:
        line = line.strip()
        if not line:
            continue
        fields = line.split()
        if len(fields) != 2:
            sys.exit("invalid input: " + repr(line))
        g.add(fields[1], fields[0])

    nodes = list(g.static_order())
    for s in nodes:
        print(s)

    if file != sys.stdin:
        file.close()



#!/usr/bin/env python3

_license = """
big
Copyright 2022 Larry Hastings
All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import itertools
import unittest

from big import TopologicalSorter


def _parse(nodes_and_dependencies):
    args = []
    for line in nodes_and_dependencies.strip().split("\n"):
        line = line.strip()
        if (not line) or line.startswith("#"):
            continue
        a = line.split()
        args.append(a)
    return args

def parse(nodes_and_dependencies):
    args = _parse(nodes_and_dependencies)
    graph = TopologicalSorter()
    for a in args:
        graph.add(*a)
    return graph

tests_run = 0

class TopologicalSortTests(unittest.TestCase):

    def permute_tests(self, nodes_and_dependencies, result, *, remove="", cycle=None):
        global tests_run

        args = _parse(nodes_and_dependencies)

        # we try every ordering of adding the nodes
        # and, if there are removals, for each of those
        #    we try every ordering of the removals
        for args in itertools.permutations(args):
            if remove:
                remove_iterator = itertools.permutations(remove)
            else:
                remove_iterator = (None,)
            for removals in remove_iterator:
                graph = TopologicalSorter()
                for a in args:
                    graph.add(*a)
                if removals:
                    for s in removals:
                        graph.remove(s)

                if cycle:
                    c = graph.cycle()
                    self.assertSetEqual(set(graph.cycle()), set(cycle), msg=f"{graph.cycle()} != {cycle}")
                    return

                self.assertFalse(graph.cycle())
                yielded = []
                while graph:
                    ready = graph.ready()
                    self.assertTrue(ready)
                    yielded.extend(ready)
                    graph.done(*ready)
                got = "".join(sorted(yielded))
                self.assertEqual(got, result, msg=f"expected {result=} is not equal to {got!r}")
                tests_run += 1


    def test_simple_cycle(self):
        self.permute_tests("A A", cycle="A", result="A")

    nodes_and_dependencies = """
        A
        B   A
        C   A
        D   B C
        X   A B C D
        E   C D X
        Y   C D X E
        """

    def test_basic_graph(self):
        self.permute_tests(self.nodes_and_dependencies, result="ABCDEXY")

    def test_with_removal(self):
        self.permute_tests(self.nodes_and_dependencies, remove="XY", result="ABCDE")

    nodes_and_dependencies_with_cycle = nodes_and_dependencies + """
        F   B
        A   F
        """
    def test_complex_cycle(self):
        self.permute_tests(self.nodes_and_dependencies_with_cycle, cycle="ABF", result="ABCDE")

    def test_complex_cycle_with_removal(self):
        self.permute_tests(self.nodes_and_dependencies_with_cycle, remove="CDX", cycle="ABF", result="ABCDE")

    def test_removals(self):
        self.permute_tests("""
            A
            B   A
            C   B A
            D   C B A
            E   D C B A
            """, remove="DBC", result="AE")

    def test_incoherence(self):
        # test view incoherence:
        # if you add an edge A-1, where 1 depends on A,
        # the view is coherent only if one of these statements is true:
        #   * 1 has not been yielded, or
        #   * A has been marked as done.
        global tests_run

        for predecessor_state in range(3):
            for successor_state in range(3):
                for delete_predecessor in (False, True):
                    g = TopologicalSorter()

                    if predecessor_state != 0:
                        g.add("A")
                    if successor_state != 0:
                        g.add("1")

                    v = g.view()
                    ready = v.ready()
                    if predecessor_state != 0:
                        self.assertIn('A', ready)
                    if successor_state != 0:
                        self.assertIn('1', ready)

                    if predecessor_state == 2:
                        v.done("A")
                    if successor_state == 2:
                        v.done("1")

                    if predecessor_state == 0:
                        g.add("A")
                    if successor_state == 0:
                        g.add("1")

                    g.add('1', "A")

                    should_be_coherent = (
                        (predecessor_state == 2)
                        or
                        (successor_state == 0)
                        )

                    try:
                        bool(v)
                        coherent = True
                    except RuntimeError:
                        coherent = False

                    # print(f"{predecessor_state=} {successor_state=} {should_be_coherent=} {coherent=}")
                    # g._default_view.print()
                    # print()

                    self.assertEqual(should_be_coherent, coherent, msg=f"test 1 {predecessor_state=} {successor_state=} {should_be_coherent=} {coherent=}")
                    tests_run += 1

                    if coherent:
                        continue

                    # now delete one of the two nodes and assert that the graph is returned to coherence
                    g.remove("A" if delete_predecessor else '1')
                    bool(v)
                    v.close()
                    tests_run += 1

    def generate_groups(self):
        # first, build a list of the groups we get from an iterator
        g = parse(self.nodes_and_dependencies)
        g_groups = []
        while g:
            r = g.ready()
            g_groups.append(r)
            g.done(*r)
        return g, g_groups

    def test_reset(self):
        global tests_run
        g, g_groups = self.generate_groups()
        # test reset()
        g.reset()
        for i, r in enumerate(g_groups, 1):
            r2 = g.ready()
            self.assertEqual(r, r2, msg=f"failed at step {i}: {r=} != {r2=}")
            g.done(*r2)
            tests_run += 1

    def test_mutation(self):
        # test mutating the graph while iterating over it.
        # we add unrelated nodes at each step while walking the graph
        # and see that they're returned in proper order.
        global tests_run
        g, g_groups = self.generate_groups()

        # adding this empty tuple lets the test work even after we run out
        # of the original nodes we added (when we add 1 and 2 right at the end).
        g_groups.append(())

        for step in range(len(g_groups)):
            g = parse(self.nodes_and_dependencies)
            i = 0
            while g:
                if i == step:
                    g.add("2", "1")
                r = g.ready()
                self.assertLessEqual(set(g_groups[i]), set(r), f"{set(g_groups[i])=} isn't <= {set(r)=}")
                if i == step:
                    self.assertIn('1', r, msg=f"1 not in {r=}")
                elif i == (step + 1):
                    self.assertIn('2', r, msg=f"2 not in {r=}")
                g.done(*r)
                i += 1
            tests_run += 1

    def test_copy(self):
        g, g_groups = self.generate_groups()
        g2 = g.copy()
        order1 = g.static_order()
        order2 = g2.static_order()
        self.assertEqual(list(order1), list(order2))

    def test_len(self):
        g, g_groups = self.generate_groups()
        self.assertEqual(len(g), 7)

    def test_empty_graph_is_false(self):
        g = TopologicalSorter()
        self.assertFalse(g)

    def test_empty_graph_is_empty(self):
        g = TopologicalSorter()
        self.assertEqual(list(g.static_order()), [])

    def test_copying_incoherent_view(self):
        g = TopologicalSorter()
        g.add('B', 'A')
        g.add('C', 'A')
        g.add('D', 'B', 'C')
        ready = g.ready()
        self.assertEqual(ready, ('A',))
        g.add('A', 'D')
        with self.assertRaises(RuntimeError):
            bool(g)
        g2 = g.copy()
        with self.assertRaises(RuntimeError):
            bool(g2)
        with self.assertRaises(RuntimeError):
            g2.ready()
        with self.assertRaises(RuntimeError):
            g2.done('A')

    def test_cycle(self):
        g = TopologicalSorter()
        g.add('B', 'A')
        g.add('C', 'A')
        g.add('D', 'B', 'C')
        self.assertEqual(g.cycle(), None)
        # coverage test, we return when the dirty bit isn't set
        self.assertEqual(g.cycle(), None)

    def test_remove(self):
        g = TopologicalSorter()
        g.add('B', 'A')
        g.add('C', 'A')
        g.add('D', 'B', 'C')
        self.assertEqual(list(g.static_order()), ['A', 'B', 'C', 'D'])
        g.remove('D')
        self.assertEqual(list(g.static_order()), ['A', 'B', 'C'])
        with self.assertRaises(ValueError):
            g.remove('Q')


    def test_close(self):
        g = TopologicalSorter()
        v = g.view()
        v.close()
        with self.assertRaises(ValueError):
            bool(v)
        with self.assertRaises(ValueError):
            bool(v.ready())
        with self.assertRaises(ValueError):
            bool(v.done('A'))
        with self.assertRaises(ValueError):
            bool(v.reset())
        with self.assertRaises(ValueError):
            bool(v.close())

    def test_print(self):
        output = ""
        def print(*a, end="\n", sep=" "):
            nonlocal output
            output += sep.join(str(o) for o in a) + end

        g, g_groups = self.generate_groups()
        v = g.view()
        v.print(print=print)
        self.assertIn('nodes', output)
        self.assertIn('ready', output)
        self.assertIn('yielded', output)
        self.assertIn('done', output)
        self.assertIn('conflict', output)

    def test_manually_constructed_graph(self):
        # this isn't a supported API
        # but this test will make coverage happy.
        g = TopologicalSorter()
        v = TopologicalSorter.View(g)
        self.assertFalse(v)
        with self.assertRaises(ValueError):
            v2 = TopologicalSorter.View({})




#
# Reusing graphlib tests
#
import big.graph
graphlib = big.graph

import sys
sys.modules['graphlib'] = graphlib

import test
from test import test_graphlib
# the API has changed and they are now invalid.
for fn_name in """
    test_calls_before_prepare
    test_prepare_multiple_times
    """.strip().split():
    delattr(test.test_graphlib.TestTopologicalSort, fn_name)

import bigtestlib

def run_tests(exit=False):
    bigtestlib.run(name="big.graph", module=__name__, permutations=lambda: tests_run)
    print("Testing big.graph using test.test_graphlib...")
    bigtestlib.run(name=None, module="test.test_graphlib")


if __name__ == "__main__": # pragma: no cover
    run_tests()
    bigtestlib.finish()
